#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2004 Free Software Foundation
#
# FILE:
# Connection.py
#
# DESCRIPTION:
#
# NOTES:
#
# $Id :$

import copy

__all__ = ['Connection']


# System Init:
#   +-- GConnections initializes
#   +-- GParser initializes:
#     +-- GDataSource loaded
#   +-- GDataSources Intialized
#     +-- GConnection.getDataObject()
#       +--
#
class Connection:
  def __init__(self, connections, name, parameters):
    self.manager = connections
    self.parameters = parameters
    self.name = name

    # Text encoding used by the database.
    # (Not all databases support this concept)
    try:
      self._encoding = parameters['encoding']
    except KeyError:
      try:
         self._encoding = gConfig('textEncoding')
      except:
        # TODO: Is this a safe default?
        self._encoding = 'iso8859-1'


  # Commit changes to the database
  def commit(self):
    pass

  # Rollback changes to the database
  def rollback(self):
    pass

  # Close the connection to the database backend
  def close(self):
    pass

  def connect(self, connectData):
    pass


  # ---------------------------------------------------------------------------
  # Return an identifier usable by the connection
  # ---------------------------------------------------------------------------

  def identifier (self, identifier):
    """
    This function transforms an identifier to be safely used by the connection.
    Descendants can override this function to do their stuff like switching
    case and so on.

    @param identifier: the identifier to be processed
    @return: string with a safe identifier
    """
    return identifier


  # ---------------------------------------------------------------------------
  # update the schema definition 
  # ---------------------------------------------------------------------------

  def updateSchema (self, definition, codeOnly = False):
    """
    This function modifies the database schema according to the given
    definition, which is a dictionary of table defininitions (@see:
    Schema.Creation.Creation), where the tablenames are the keys.

    @param definition: sequence of table definitions
    @param codeOnly: if TRUE, only the code will be generated to modify the
        schema, no actions take place.
    @return: a tuple of three sequences (prologue, body, epilogue) holding the
        code to perform all schema modifications needed. These sequences should
        be executed in this order to successfully create the schema.
    """

    result = ([], [], [])

    if self.schemaCreator is None:
      return result

    workingSet    = copy.copy (definition)
    constraintSet = {}

    # before any actions are performed, validate the given definitions
    for table in definition:
      self.schemaCreator.validate (table)

    # in a first run we remove all constraint definitions from the tables, so
    # we can add or alter all tables without having troubles with order of
    # occurence.
    for table in workingSet:
      # Do we have already a table with that name?
      res = self.introspector.find (name = self.identifier (table ['name']))

      if res is not None:
        method = self.schemaCreator.modifyTable
        existingFields = [f.name for f in res [0].fields ()]

        # keep only new fields
        keep = []
        for field in table ['fields']:
          if not self.identifier (field ['name']) in existingFields:
            keep.append (field)

        table ['fields'] = keep

        # on updates of a table we cannot use a primary key
        if table.has_key ('primarykey'):
          del table ['primarykey']

        # we create an index on a table update only if it contains new fields
        if table.has_key ('indices'):
          keep = []
          for index in table ['indices']:
            for field in index ['fields']:
              if not self.identifier (field) in existingFields:
                keep.append (index)
                break

          table ['indices'] = keep

        # we create a constraint on a table update only if it contains new
        # fields
        if table.has_key ('constraints'):
          keep = []

          for constraint in table ['constraints']:
            for field in constraint ['fields']:
              if not self.identifier (field) in existingFields:
                keep.append (constraint)
                break

          table ['constraints'] = keep

      else:
        method = self.schemaCreator.createTable

      if table.has_key ('constraints'):
        constraintSet [table ['name']] = {'name':        table ['name'],
                                          'constraints': table ['constraints']}
        del table ['constraints']

      # before we execute the planned action, have a look if there's still work
      # to be done
      perform = table.has_key ('fields') and len (table ['fields'])
      perform = perform or table.has_key ('primarykey')
      perform = perform or (table.has_key ('indices') and \
                            len (table ['indices']))

      if perform:
        self.schemaCreator.mergeTuple (result, method (table, codeOnly))

    # on the second run we process all constraints
    for table in constraintSet.values ():
      self.schemaCreator.mergeTuple (result,
                              self.schemaCreator.createTable (table, codeOnly))

    return result
