#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or(at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2000-2004 Free Software Foundation
#
# $Id: Introspection.py 5944 2004-07-06 13:19:38Z johannes $

__all__ = ['Introspection']

import string
import re

from gnue.common.datasources import GIntrospection

# =============================================================================
# This class implements schema introspection for Interbase / Firebird
# =============================================================================

class Introspection(GIntrospection.Introspection):

  # list of the types of Schema objects this driver provides
  types = [('table', _('Tables'), 1),
           ('view' , _('Views') , 1)]

  _NOW = re.compile ("'(NOW\s*\(\)\s*)'", re.IGNORECASE)


  # ---------------------------------------------------------------------------
  # Find a schema element by name and/or type
  # ---------------------------------------------------------------------------

  def find (self, name = None, type = None):
    """
    This function searches the schema for an element by name and/or type. If no
    name and no type is given, all elements will be retrieved.

    @param name: look for an element with this name
    @param type: look for an element with this type
    @return: A sequence of schema instances, one per element found, or None if
        no element could be found.
    """

    result = []
    cond   = ["rdb$system_flag = 0"]

    if name is not None:
      cond.append (u"rdb$relation_name = '%s'" \
                   % self._connection.identifier (name))

    if type == 'table':
      cond.append (u"rdb$view_source IS NULL")
    elif type == 'view':
      cond.append (u"rdb$view_source IS NOT NULL")

    cmd = u"SELECT rdb$relation_name, rdb$view_source FROM RDB$RELATIONS " \
           "WHERE %s ORDER BY rdb$relation_name" % string.join (cond, " AND ")

    cursor = self._connection.makecursor (cmd)

    try:
      for rs in cursor.fetchall ():
        relname = self._connection.identifier (string.strip (rs [0]))

        attrs = {'id'        : relname,
                 'name'      : relname,
                 'type'      : rs [1] is None and 'table' or 'view',
                 'primarykey': self.__getPrimaryKey (relname)}

        result.append ( \
          GIntrospection.Schema (attrs, getChildSchema = self._getChildSchema))

    finally:
      cursor.close ()

    return len (result) and result or None


  # ---------------------------------------------------------------------------
  # Get all fields of a relation / view
  # ---------------------------------------------------------------------------

  def _getChildSchema (self, parent):
    """
    This function returns a list of all child elements of a given parent
    relation.

    @param parent: schema object instance whose child elements should be
        fetched.
    @return: sequence of schema instances, one per element found
    """

    result = []

    cmd = u"SELECT rf.rdb$field_name, tp.rdb$type_name, rf.rdb$null_flag, " \
             "rf.rdb$default_source, fs.rdb$field_length, " \
             "fs.rdb$field_scale, fs.rdb$field_precision " \
           "FROM rdb$relation_fields rf, rdb$fields fs, rdb$types tp " \
           "WHERE rf.rdb$relation_name = '%s' AND " \
             "fs.rdb$field_name = rf.rdb$field_source AND " \
             "tp.rdb$type = fs.rdb$field_type AND " \
             "tp.rdb$field_name = 'RDB$FIELD_TYPE'" \
           "ORDER BY rf.rdb$field_name" % parent.name

    cursor = self._connection.makecursor (cmd)

    try:
      for rs in cursor.fetchall ():
        nativetype = rs [1].strip ()
        attrs = {'id'        : "%s.%s" % (parent.name, rs [0].strip ()),
                 'name'      : self._connection.identifier (rs [0].strip ()),
                 'type'      : 'field',
                 'nativetype': nativetype,
                 'required'  : rs [2] is not None}

        if nativetype in ['DATE', 'TIME', 'TIMESTAMP']:
          attrs ['datatype'] = 'date'

        elif nativetype in ['DOUBLE', 'FLOAT', 'INT64', 'LONG', 'QUAD', \
            'SHORT']:
          attrs ['datatype']  = 'number'
          attrs ['length']    = rs [6]
          attrs ['precision'] = abs (rs [5])

        else:
          attrs ['datatype'] = 'text'
          attrs ['length'] = rs [4]

        if rs [3] is not None:
          default = rs [3]
          if self._NOW.search (default) is not None:
            attrs ['defaulttype'] = 'timestamp'
          else:
            attrs ['defaulttype'] = 'constant'
            attrs ['defaultval']  = default [8:]

        result.append (GIntrospection.Schema (attrs))

    finally:
      cursor.close ()

    return result


  # ---------------------------------------------------------------------------
  # Get all parts of a primary key
  # ---------------------------------------------------------------------------

  def __getPrimaryKey (self, relname):
    """
    This function fetches all fields of a primary key for the given relation.

    @param relid: object id of the relation in question
    @return: sequence with the fieldnames of the primary key or None, if the
        relation has no primary key.
    """

    cmd = u"SELECT rdb$field_name " \
           "FROM rdb$relation_constraints rc, rdb$index_segments ri " \
           "WHERE ri.rdb$index_name = rc.rdb$index_name " \
              "AND rc.rdb$constraint_type = 'PRIMARY KEY' " \
              "AND rc.rdb$relation_name = '%s'" \
           "ORDER BY ri.rdb$field_position" % relname

    cursor = self._connection.makecursor (cmd)
    
    try:
      result = [rs [0].strip () for rs in cursor.fetchall ()]

    finally:
      cursor.close ()

    return len (result) and result or None
