#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001-2004 Free Software Foundation
#
# $Id: Creation.py 5920 2004-06-21 18:11:52Z johannes $

from gnue.common.datasources.drivers.DBSIG2.Schema.Creation import \
    Creation as Base


# =============================================================================
# Class implementing schema creation for MySQL (3.x/4.x)
# =============================================================================

class Creation (Base.Creation):

  MAX_NAME_LENGTH = 64

  # ---------------------------------------------------------------------------
  # Handle special defaults
  # ---------------------------------------------------------------------------

  def _defaultwith (self, code, tableName, fieldDef, forAlter):
    """
    This function adds 'auto_increment' for 'serials' and checks for the proper
    fieldtype on 'timestamps'

    @param code: code-tuple to merge the result in
    @param tableName: name of the table
    @param fieldDef: dictionary describing the field with the default
    @param forAlter: TRUE if the definition is used in a table modification
    """
    if fieldDef ['defaultwith'] == 'serial':
      seq = self._getSequenceName (tableName, fieldDef)
      code [1] [-1] += " AUTO_INCREMENT"
      fieldDef ['default'] = "nextval ('%s')" % seq

    elif fieldDef ['defaultwith'] == 'timestamp':
      if fieldDef ['type'] != 'timestamp':
        fieldDef ['type'] = 'timestamp'

        code [1].pop ()
        code [1].append (self._composeField (tableName, fieldDef, forAlter))
        
        print _("WARNING: changing column type of '%(table)s.%(column)s' "
                "to 'timestamp'") \
              % {'table': tableName,
                 'column': fieldDef ['name']}


  # ---------------------------------------------------------------------------
  # A key is an unsigned integer
  # ---------------------------------------------------------------------------

  def key (self, fieldDefinition):
    """
    Native datatype for a 'key'-field is 'unsigned integer'

    @param fieldDefinition: dictionary describing the field
    @return: 'int unsigned'
    """
    return "int unsigned"


  # ---------------------------------------------------------------------------
  # Translate a string into an apropriate native type
  # ---------------------------------------------------------------------------

  def string (self, fieldDefinition):
    """
    This function returns an apropriate native type for a string. If the length
    is given and below 255 character the result is a varchar, otherwist text.

    @param fieldDefinition: dictionary describing the field
    @return: string with the native datatype
    """
    if fieldDefinition.has_key ('length') and fieldDefinition ['length'] <= 255:
      return "varchar (%s)" % fieldDefinition ['length']
    else:
      return "text"


  # ---------------------------------------------------------------------------
  # Create an apropriate type for a number
  # ---------------------------------------------------------------------------

  def number (self, fieldDefinition):
    """
    This function returns an apropriate type for a number according to the
    given length and precision.

    @param fieldDefinition: dictionary describing the field
    @return: string with the native datatype
    """
    scale  = 0
    length = 0

    if fieldDefinition.has_key ('precision'):
      scale = fieldDefinition ['precision']
    if fieldDefinition.has_key ('length'):
      length = fieldDefinition ['length']

    if scale == 0:
      if length <= 4:
        return "smallint"
      elif length <= 9:
        return "int"
      elif length <= 18:
        return "bigint"
      else:
        return "decimal (%s,0)" % length
    else:
      return "decimal (%s,%s)" % (length, scale)


  # ---------------------------------------------------------------------------
  # MySQL has no native boolean type
  # ---------------------------------------------------------------------------

  def boolean (self, fieldDefinition):
    """
    MySQL has no native boolean type, so this function returns 'tinyint (1)
    unsigned' instead.

    @param fieldDefinition: dictionary describing the field
    @return: 'tinyint (1) unsigned'
    """
    return "tinyint (1) unsigned"


  # ---------------------------------------------------------------------------
  # MySQL has a timestamp, which is needed for 'defaultwith timestamp'
  # ---------------------------------------------------------------------------

  def timestamp (self, fieldDefinition):
    """
    In MySQL timestamps are used for default values, otherwise we map to the
    inherited typetransformation of 'timestamp'.

    @param fieldDefinition: dictionary describing the field
    @return: string with the native datatype
    """
    if fieldDefinition.has_key ('defaultwith') and \
        fieldDefinition ['defaultwith'] == 'timestamp':
      return "timestamp"
    else:
      return Base.Creation.timestamp (self, fieldDefinition)

