#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2001-2004 Free Software Foundation
#
# $Id: language.py 5822 2004-05-21 09:37:07Z johannes $

import string
from gnue.common.utils.FileUtils import dyn_import

adapters = {}


# =============================================================================
# Exceptions
# =============================================================================

# ---------------------------------------------------------------------------
# The basis of all language errors
# ---------------------------------------------------------------------------
class Error (gException):
  pass

# -----------------------------------------------------------------------------
# Python module not available
# -----------------------------------------------------------------------------

class AdapterNotFoundError (Error):
  """
  The language adapter for the requested language cannot be imported.
  """
  def __init__ (self, language):
    msg = u_("No adapter available for language '%s'") % language
    Error.__init__ (self, msg)


# -----------------------------------------------------------------------------
# Abstract method not implemented
# -----------------------------------------------------------------------------

class ImplementationError (Error):
  """
  Exception raised if an abstract method isn't implemented by a descendant.
  """
  def __init__ (self, classname, method):
    msg = u_("The class '%(class)s' has no implementation for '%(method)s'") \
           % {"class" : classname,
              "method": method}
    Error.__init__ (self, msg)

# -----------------------------------------------------------------------------
# Code failed on compilation
# -----------------------------------------------------------------------------

class CompileError (Error):
  pass

# -----------------------------------------------------------------------------
# Code failed on execution
# -----------------------------------------------------------------------------

class RuntimeError (Error):
  """
  Indicates an exception that happend on execution of a method/function.

  If e is an exception of this type, then e.message is the message text of the
  exception that occured, e.type is a string containing the name of the type of
  the exception, and e.detail is the complete traceback.
  """
  def __init__ (self, traceback):
    self.detail = traceback
    self.type   = ''
    message     = ''

    # Search for start of Exception type/value part
    list = string.split (traceback, '\n')
    for i in range (1, len (list) - 1):
      if len (list [i]) > 0 and list [i] [0] != ' ':
        info = string.split (list [i], ': ', 1)
        if len (info) == 2:
          self.type = info [0]
          message = string.join ([info [1]] + list [i+1:], '\n')
          break
    Error.__init__ (self, message)


# -----------------------------------------------------------------------------
# Abort the current execution
# -----------------------------------------------------------------------------

class AbortRequest (Error):
  pass

# -----------------------------------------------------------------------------
# Get or create an instance of a given language adapter
# -----------------------------------------------------------------------------

def getLanguageAdapter (language):
  """
  This function returns an execution context factory for the given language.
  AdapterNotFoundError will be raised, if the language adapter cannot be
  imported.
  """
  global adapters
  lang = language.lower ()

  if not adapters.has_key (lang):
    try:
      adapter = dyn_import ('gnue.common.logic.adapters.%s' % lang)
      adapters [lang] = adapter.LanguageAdapter ()

    except ImportError:
      raise AdapterNotFoundError, language

  return adapters [language]


# =============================================================================
# Self test code
# =============================================================================

if __name__ == '__main__':

  print "Creating language adapter for 'python' ..."
  adapter = getLanguageAdapter ('python')

  print "Creating new execution environment ..."
  environ = adapter.createNewContext ()
  environ.shortname   = "testing"
  environ.description = "Execution context for self testing"

  code = """
print "Hello World!"
print "My name is %s." % name
print fuck
return value * 2
"""

  print "Creating a new virtual code object ..."
  method = environ.buildFunction ('myFunctionName', code,
                        {'name': "", 'value': 0})

  params = {'name': 'foo', 'value': 'bar'}
  print "Calling function with: %s" % params
  res = method (**params)
  print "   result:", repr (res)

  params = {'name': 'fooBar', 'value': 4}
  print "Calling function with: %s" % params
  res = method (**params)
  print "   result:", repr (res)

  print "Thank's for playing."
