#
# This file is part of GNU Enterprise.
#
# GNU Enterprise is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2, or (at your option) any later version.
#
# GNU Enterprise is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with program; see the file COPYING. If not,
# write to the Free Software Foundation, Inc., 59 Temple Place
# - Suite 330, Boston, MA 02111-1307, USA.
#
# Copyright 2002-2004 Free Software Foundation
#
# $Id: Scripter.py 6119 2004-08-02 08:48:32Z johannes $

from gnue.common import VERSION
from gnue.common.schema import GSParser
from gnue.common.utils.FileUtils import openResource
from gnue.common.apps.GClientApp import GClientApp
from gnue.common.datasources import GDataSource, GConditions

from time import strftime
from string import join

import sys
import os
import re
import copy
import types


# =============================================================================
# Exceptions
# =============================================================================

class Error (gException):
  pass

class MissingKeyError (Error):
  def __init__ (self, tableName):
    msg = u_("Data row of table '%s' has no key fields") % tableName
    Error.__init__ (self, msg)

# =============================================================================
# Load GNUe Schema Definition files and create a database schema for it
# =============================================================================

class Scripter (GClientApp):

  VERSION         = VERSION
  COMMAND         = "gnue-schema"
  NAME            = "GNUe Schema Scripter"
  USAGE           = "[options] gsd-file [gsd-file gsd-file ...]"
  SUMMARY = _("GNUe Schema Scripter creates database schemas based on GNUe "
              "Schema Definitions.")


  # ---------------------------------------------------------------------------
  # Constructor
  # ---------------------------------------------------------------------------

  def __init__ (self, connections = None):
    self.addCommandOption ('connection', 'c', argument='connectionname',
        help = _("Use the connection <connectionname> for creating the schema"))

    self.addCommandOption ('output','o', argument='filename',
        help = _("Also send the code for creating the schema to this file."))

    self.addCommandOption ('file-only', 'f', default = False,
        help = _("If this flag is set, only code is sent to the output file "
                 "and the schema is not created automatically."))

    self.addCommandOption ('mode', 'm', argument='both|schema|data',
        default = 'both',
        help = _("Mode of operation. If mode is 'schema', only schema "
                 "creation is done. If mode is 'data' only data integration "
                 "is done."))

    self.addCommandOption ('username', 'u', argument="user",
        help = _("Set the username for the database. If the database is to be "
                 "created, this username will be it's owner."))

    self.addCommandOption ('password', 'p', argument="password",
        help = _("Set the password for the database."))

    self.addCommandOption ('createdb', 'd', default = False,
        help = _("If this option is set, the database will be created before "
                 "any schema creation is done. There must be a username "
                 "either from the given connection-configuration or from the "
                 "command line. This user becomes the owner of the database "
                 "and will be implicitly created."))

    ConfigOptions = {}
    GClientApp.__init__ (self, connections, 'schema', ConfigOptions)



  # ---------------------------------------------------------------------------
  # Main program
  # ---------------------------------------------------------------------------

  def run (self):
    """
    This is the main function of the whole process. It verifies the given
    options, loads all schema definitions and then logs into the connection to
    perform all actions requested.
    """

    self.__checkOptions ()

    try:
      self.tables    = []
      self.tabledata = []

      for item in range (len (self._files)):
        print o (u_("Loading gsd file '%s' ...") % self.ARGUMENTS [item])

        try:
          schema = GSParser.loadFile (self._files [item])
          schema.walk (self.__iterateObjects)

        finally:
          self._files [item].close ()

    except Exception:
      print sys.exc_info () [1]

    else:
      try:
        if self.__doData and len (self.tabledata):
          self.verifyDataKeys ()

        if self.OPTIONS ['createdb']:
          if self.connection.schemaCreator is not None:
            self.connection.schemaCreator.createDatabase ()

        if self.__doSchema:
          self.executeAndGenerateCode ()

        if self.__doData and len (self.tabledata):
          self.updateData ()

      except Exception:
        print sys.exc_info () [1]



  # ---------------------------------------------------------------------------
  # Walk through all command line options
  # ---------------------------------------------------------------------------

  def __checkOptions (self):
    """
    This function checks wether the given command line arguments and options
    are usefull or not.
    """

    if not len (self.ARGUMENTS):
      self.handleStartupError (_("No input file specified."))

    try:
      self._files = []

      for filename in self.ARGUMENTS:
        self._files.append (openResource (filename))

    except IOError:
      self.handleStartupError (_("Unable to open input file %s.") % filename)


    if not self.OPTIONS ['connection']:
      self.handleStartupError (_("No connection specified."))

    self.outfile = self.OPTIONS ['output']

    if self.OPTIONS ['file-only'] and self.outfile is None:
      self.handleStartupError (_("Output to file only requested, but no "
                                 "filename specified."))

    self.__doSchema = self.OPTIONS ['mode'].lower () in ['both', 'schema']
    self.__doData   = self.OPTIONS ['mode'].lower () in ['both', 'data'] and \
                          not self.OPTIONS ['file-only']

    if not (self.__doSchema or self.__doData):
      self.handleStartupError (_("Mode of operation must be one of "
                                 "'both', 'schema' or 'data'."))

    cName = self.OPTIONS ['connection']
    self.connection = self.connections.getConnection (cName)

    # if a username is given on the command line we pass both username and
    # password to the connection parameters. If the password is not set, it
    # defaults to an empty string.
    username = self.OPTIONS ['username'] or 'gnue'
    password = self.OPTIONS ['password'] or 'gnue'

    self.connection.parameters ['username'] = username
    self.connection.parameters ['password'] = password



  # ---------------------------------------------------------------------------
  # Get a dictionary with all keys listed in tags and values from sObject
  # ---------------------------------------------------------------------------

  def fetchTags (self, sObject, tags):
    """
    This function creates a dictionary with all attributes from sObject listed
    in tags, where the attributenames are the keys.

    @param sObject: Schema object to retriev attributes from
    @param tags: list of all attributes to retrieve
    @return: dictionary with the attribute names as keys and their values
    """
    res = {}
    for item in tags:
      if hasattr (sObject, item):
        res [item] = getattr (sObject, item)
    return res


  # ---------------------------------------------------------------------------
  # iteration over all schema objects in the document tree
  # ---------------------------------------------------------------------------

  def __iterateObjects (self, sObject):
    """
    This is the master iteration function, it runs over all top level objects
    in the GSObject tree and processes GSTable- and GSTableData-subtrees.

    @param sObject: current Schema object to be processed
    """
    if sObject._type == "GSTable":
      self.tables.append ({'name': sObject.name})
      sObject.walk (self.__schemaFields, defs = self.tables [-1])

    elif sObject._type == "GSTableData":
      self.tabledata.append ({'name': sObject.tablename, 'rows': []})
      sObject.walk (self.__dataRows, defs = self.tabledata [-1]['rows'])

    return


  # ---------------------------------------------------------------------------
  # Process the fields of a GSTable
  # ---------------------------------------------------------------------------

  def __schemaFields (self, sObject, defs):
    """
    This function iterates over all child elements of a GSTable instance and
    converts this subtree into a table definition dictionary given by the
    parameter defs.

    @param sObject: current schema object to be processed
    @param defs: dictionary of the table definition, which has to be extended
        by this function.
    """

    # process a regular field of a table
    if sObject._type == "GSField":
      fDef = self.fetchTags (sObject, ['name', 'type', 'default',
                        'defaultwith', 'length', 'precision', 'nullable'])
      if not defs.has_key ('fields'):
        defs ['fields'] = []
      defs ['fields'].append (fDef)

    # add a primary key definition to the table definition
    elif sObject._type == "GSPrimaryKey":
      pkDef = {'name': sObject.name, 'fields': []}
      defs ['primarykey'] = pkDef
      sObject.walk (self.__schemaPrimaryKey, defs = pkDef)

    # start an index definition and process it's fields
    elif sObject._type == "GSIndex":
      if not defs.has_key ('indices'):
        defs ['indices'] = []

      indexDef = {'name': sObject.name,
                  'unique': hasattr (sObject, "unique") and sObject.unique,
                  'fields': []}
      defs ['indices'].append (indexDef)

      sObject.walk (self.__schemaIndex, defs = indexDef)

    # create constraints
    elif sObject._type == "GSConstraint":
      # for unique-constraints we use a 'unique index'
      if sObject.type == "unique":
        if not defs.has_key ('indices'):
          defs ['indices'] = []

        cDef = {'name'  : sObject.name,
                'unique': True,
                'fields': []}
        defs ['indices'].append (cDef)

      # for all other types of constraints we use a ConstraintDefinition
      else:
        if not defs.has_key ('constraints'):
          defs ['constraints'] = []

        cDef = self.fetchTags (sObject, ['name', 'type'])
        cDef ['fields'] = []
        defs ['constraints'].append (cDef)

      sObject.walk (self.__schemaConstraint, defs = cDef)


  # ---------------------------------------------------------------------------
  # Iterate over all fields of a primary key
  # ---------------------------------------------------------------------------

  def __schemaPrimaryKey (self, sObject, defs):
    """
    This function converts all GSPKField instances into a primary key
    definition (dictionary).

    @param sObject: current schema object to be processed
    @param defs: dictionary describing the primary key
    """
    if sObject._type == "GSPKField":
      defs ['fields'].append (sObject.name)


  # ---------------------------------------------------------------------------
  # Iterate over all fields of an index
  # ---------------------------------------------------------------------------

  def __schemaIndex (self, sObject, defs):
    """
    This function converts all GSIndexField instances into a index definition
    (dictionary).

    @param sObject: current schema object to be processed
    @param defs: dictionary describing the index
    """
    if sObject._type == "GSIndexField":
      defs ['fields'].append (sObject.name)


  # ---------------------------------------------------------------------------
  # Iterate over all children of a constraint definition
  # ---------------------------------------------------------------------------

  def __schemaConstraint (self, sObject, defs):
    """
    This function converts constraint fields and references into a constraint
    definition (dictionary).

    @param sObject: current schema object to be processed
    @param defs: dictionary describing the constraint
    """
    if sObject._type == "GSConstraintField":
      defs ['fields'].append (sObject.name)

    elif sObject._type == "GSConstraintRef":
      defs ['reftable'] = sObject.table
      if not defs.has_key ('reffields'):
        defs ['reffields'] = []
      defs ['reffields'].append (sObject.name)


  # ---------------------------------------------------------------------------
  # Iterate over all rows of a tabledata definition
  # ---------------------------------------------------------------------------

  def __dataRows (self, sObject, defs):
    """
    This function converts a GSRow instance into an element in the given data
    dictionary.

    @param sObject: current schema object to be processed
    @param defs: sequence of row definitions of the corresponding table
    """
    if sObject._type == "GSRow":
      defs.append ({'key': [], 'fields': []})
      sObject.walk (self.__dataValues, defs [-1])


  # ---------------------------------------------------------------------------
  # Iterate over all values of a row definition
  # ---------------------------------------------------------------------------
  def __dataValues (self, sObject, defs):
    """
    This function translates a GSValue instance into an element of a row
    definition dictionary.

    @param sObject: current schema object to be processed.
    @param defs: row definition dictionary to be extended.
    """
    if sObject._type == "GSValue":
      defs ['fields'].append ({'name' : sObject.field, 'value': sObject.value})
      if hasattr (sObject, "key") and sObject.key:
        defs ['key'].append (sObject.field)


  # ---------------------------------------------------------------------------
  # Make sure all rows in the table data sequence have a valid key
  # ---------------------------------------------------------------------------

  def verifyDataKeys (self):
    """
    This function iterates over all data rows and makes sure they all have an
    appropriate key specified. Given all data rows of a table, there must be at
    least on row with the key-attribute set on it's fields. If there is no
    row with key information, but a table definition with a primary key this
    key will be used.

    @raise MissingKeyError: If no key information could be found for all rows
        of a table this exception will be raised.
    """
    for item in self.tabledata:
      tableName = item ['name']
      rows      = item ['rows']
      tableKey  = None
      missing   = False

      # is a key missing or does at least one row specify a key?
      for row in rows:
        if len (row ['key']):
          tableKey = row ['key']
        else:
          missing = True

      if not missing:
        continue

      # is there a table definition with a valid primary key available ?
      if tableKey is None:
        for item in self.tables:
          if item ['name'] == tableName and item.has_key ('primarykey'):
            tableKey = item['primarykey']['fields']

      # if at least one key was available, fill up all missing keys with this
      # one.
      if tableKey is not None:
        for row in rows:
          if not len (row ['key']):
            row ['key'] = tableKey
        continue

      raise MissingKeyError, (tableName)


  # ---------------------------------------------------------------------------
  # Execute and generate the code
  # ---------------------------------------------------------------------------

  def executeAndGenerateCode (self):
    """
    This function logs into the given connection and calls it for an update of
    it's schema according to the loaded table definitions. Additionally the
    schema creation code is generated by this call, which will be stored in the
    given output file (if requested by options).
    """

    self.connections.loginToConnection (self.connection)

    print _("Updating schema ...")
    code = self.connection.updateSchema (self.tables,
                                         self.OPTIONS ['file-only'])
    if self.outfile is not None:
      dest = open (self.outfile, 'w')

      for item in code:
        for line in item:
          dest.write (line + "\n")

      dest.close ()


  # ---------------------------------------------------------------------------
  # Update connection with table data 
  # ---------------------------------------------------------------------------

  def updateData (self):
    """
    This function updates the backend with data. Every row is checked using the
    specified key-fields and updated if it exists otherwise inserted.
    """

    print _("Updating data ...")
      
    for table in self.tabledata:
      tablename = table ['name']
      rows      = table ['rows']

      if not len (rows):
        continue

      attributes = {'name'    : "dts_%s" % tablename,
                    'database': self.OPTIONS ['connection'],
                    'table'   : tablename}

      fieldList = [field ['name'] for field in rows [0]['fields']]

      datasource = GDataSource.DataSourceWrapper (
          connections = self.connections,
          attributes  = attributes,
          fields      = fieldList,
          unicodeMode = True)

      print o (u_("  updating table '%s' ...") % tablename)
      updCount = 0
      insCount = 0

      for row in rows:
        cList = ['and']

        for keyField in row ['key']:
          for field in row ['fields']:
            if field ['name'] == keyField:
              kvalue = field ['value']
              break

          cList.append (['eq', ['field', keyField], ['const', kvalue]])

        condition = GConditions.buildTreeFromList (cList)
        resultSet = datasource.createResultSet (condition)

        if resultSet.firstRecord () is None:
          resultSet.insertRecord ()
          insCount += 1
        else:
          updCount += 1

        for field in row ['fields']:
          resultSet.current.setField (field ['name'], field ['value'])

        resultSet.post ()

      self.connections.commitAll ()
      print o (u_("    Inserted %d row(s), updated %d row(s)") \
               % (insCount, updCount))



# =============================================================================
# If executed directly, start the scripter
# =============================================================================
if __name__ == '__main__':
  Scripter ().run ()
